#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "main.h"
#include "globals.h"
#include "misc.h"
#include "game.h"
#include "test.h"

void game(void) {
    static state_t state=idle;
    static int cnt, level, difficulty;

    switch(state) {
        case idle:      // run a light show and wait for a touch
            enable_light_show = 1;
            if ( count_touch(matrix) ) {
                enable_light_show = 0;
                state = first_touch;
            }
            //print_pattern(matrix);
            break;

        case first_touch:   // wait for the touch to end
            if ( count_touch(matrix) == 0 ) {
                set_all_LEDs(leds, (PGM_P)menu_select_difficulty);
                update_LEDs(leds);
                state = init;
            }

        case init:      // check touch for difficulty selection, init game
            difficulty = check_menu_select(matrix);
            if ( difficulty < 3 ) {
                memset(leds, 0, sizeof(leds));
                update_LEDs(leds);
                cnt = 0;
                state = prepare;                 
            }
            break;

        case prepare:
            cnt++;
            // blink selected LED for 2s
            if (cnt < 20) {
                if (cnt & 1) {
                    switch (difficulty) {
                        case 0: set_single_LED(leds, 7, GREEN);  break;
                        case 1: set_single_LED(leds, 0, YELLOW); break;
                        case 2: set_single_LED(leds, 1, RED);    break;
                    }
                } else {
                    memset(leds, 0, sizeof(leds));                    
                }
                update_LEDs(leds);
            } else {    // let's go
                level = 0;
                generate_pattern(difficulty, level, pattern, leds);
                update_LEDs(leds);
                cnt = pgm_read_byte(&levels[difficulty][level][2]);              
                state = run;
            }
            break;

        case run:      // game running
            cnt--;
            if (compare_matrix(matrix, pattern) ) {
                cnt = 0;
                state = next_level;
            } else if (cnt == 0) {
#ifdef ALWAYS_WIN
#warning "Cheater switch ALWAYS_WIN active, use for debug only!"
                cnt = 0;
                state = next_level;       // for TEST only

#else
                game_over_light(1);
                cnt = 30;
                state = game_over;
#endif               
            }
            break;

        case next_level:
            if (cnt < 5) {
                cnt++;
            } else {
                // match
                level++;
                if (level == GAME_ROUNDS) {
                    cnt = 30;
                    victory_light(1);
                    state = victory;
                } else {
                    // next round
                    generate_pattern(difficulty, level, pattern, leds);
                    update_LEDs(leds);
                    cnt = pgm_read_byte(&levels[difficulty][level][2]);
                    state = run;
                }
            }
            break;

        case victory:   // itself
            cnt--;
            victory_light(0);
            if (!cnt) {
                cnt = 30;
                state = delay;
            }
            break;

        case game_over: // itself
            cnt--;
            game_over_light(0);
            if (!cnt) {
                cnt = 30;
                state = delay;
            }
            break;

        case delay:     // a little delay after game over/victory
            cnt--;
            if (!cnt) {
                state = idle;
            }
            break;
    }
}

// make a light show 

void light_show(void) {
    static int cnt[8] = {0, 192, 384, 576, 768, 960, 1152, 1344};

    int i, j, k, dir;

    for (i=0; i<8; i++) {
        cnt[i] +=20;
        if (cnt[i] > 1535) cnt[i]=0;
        j = cnt[i] >> 8;
        k = pgm_read_byte(&color_circle[j][3]);
        if (k>2) {
            dir = 1;
            k  -= 3;
        } else {
            dir = -1;
        }            
        leds[i][0]  = pgm_read_byte(&color_circle[j][0]);
        leds[i][1]  = pgm_read_byte(&color_circle[j][1]);
        leds[i][2]  = pgm_read_byte(&color_circle[j][2]);
        leds[i][k] += (cnt[i] & 0xFF)*dir; 
    }

    update_LEDs(leds);
}

void game_over_light(int reset) {
    int i;
    static int cnt;

    if (reset) {
        cnt = 0;
    } else {
        cnt++;
        if (cnt >= 2) cnt=0;
    }

    if (cnt < 1) {
        for (i=0; i<8; i++) {
            leds[i][1]=0xFF;    // red
        }
    } else if (cnt < 2 ) { 
        memset(leds, 0, sizeof(leds));  // all black
    }

    update_LEDs(leds);
}

void victory_light(int reset) {
    int i;
    static int cnt;

    if (reset) {
        cnt = 0;
    } else {
        cnt++;
        if (cnt >=2) cnt = 0;
    }

    if (cnt < 1) {
        for (i=0; i<8; i++) {
            leds[i][0]=0xFF;    // green
        }
    } else if (cnt < 2 ) { 
        memset(leds, 0, sizeof(leds));  // all black
    } 

    update_LEDs(leds);
}

// generate a new pattern for the game

void generate_pattern(int difficulty, int level, matrix_t set_pattern, led_t led_data) {

    uint8_t new_colors[5];
    uint8_t new_pattern;
    uint8_t new_shift=0;
    uint8_t i, j, k, l, m, n, min, max;
    uint8_t is_same, is_allowed;
    static uint8_t old_pattern, old_shift;

    // generate random pattern, colors and phase shift

    new_shift = rand_range(0, 7);
    min = pgm_read_byte(&levels[difficulty][level][0]);
    max = pgm_read_byte(&levels[difficulty][level][1]);
    new_pattern = rand_range(min, max);

    if (new_pattern == old_pattern ) {   // avoid same shift when pattern is same
        // check if new shift is allowed, avoid rotated patterns which will yield same touch connections 
        do {
            new_shift = rand_range(0, 7);
            k = (new_shift - old_shift) & 7;
            is_allowed = 1;
            for (i=0; i<4; i++) {
                j = pgm_read_byte(&patterns[new_pattern].delta[i]);
                if (j==k) {
                    is_allowed = 0;
                    break;
                }
                if (j>7) break;    // no more invalid deltas in list 
            }
        } while (!is_allowed);
    }
    old_pattern = new_pattern;
    old_shift = new_shift;

    new_colors[0] = 0;
    for (i=1; i<5; i++) {
        // new colors must be all different!
        do {
            k = rand_range(1, 6);
            is_same = 0;
            // check all previously generated colors
            for (j=1; j<i; j++) {    
                if (k == new_colors[j]) {
                    is_same=1;
                }
            }
        } while (is_same);
        new_colors[i] = k;
    }

#ifdef PRINT_DEBUG

    char tmp[80];

    soft_uart_puts_P(PSTR("DEBUG PRINT\r\n"));
    sprintf_P(tmp, PSTR("shift: %d\r\n"), new_shift );
    soft_uart_puts(tmp);
    sprintf_P(tmp, PSTR("min: %d\r\n"), min );
    soft_uart_puts(tmp);
    sprintf_P(tmp, PSTR("max: %d\r\n"), max );
    soft_uart_puts(tmp);
    sprintf_P(tmp, PSTR("pattern: %d\r\n"), new_pattern );
    soft_uart_puts(tmp);
    sprintf_P(tmp, PSTR("colors: ") );
    soft_uart_puts(tmp);

    for(i=0; i<5; i++) {
        sprintf_P(tmp, PSTR("%d "), new_colors[i]);
        soft_uart_puts(tmp);
    }
    soft_uart_puts_P(PSTR("\r\n"));

#endif

    // copy new pattern to LEDs

    for (i=0; i<8; i++) {
        k = (i+new_shift) & 0x7;    // phase shift of target index with wrap around
        j = new_colors[pgm_read_byte(&patterns[new_pattern].pattern[i])];
        led_data[k][0] = pgm_read_byte(&colors[j][0]);
        led_data[k][1] = pgm_read_byte(&colors[j][1]);
        led_data[k][2] = pgm_read_byte(&colors[j][2]);
    }

    // clear set_pattern

    memset(set_pattern, 0, sizeof(matrix_t));

    // scan pattern and set connections in set_pattern matrix
    for (j=1; j<=4; j++) {   // scan max. 4 groups
        for (n=1; n<=3; n++) {   // max. 3 members per group
            // search n-th occurence of group
            for (i=0, m=0; i<7; i++) {
                if (pgm_read_byte(&patterns[new_pattern].pattern[i]) == j) {
                    m++;
                    if (m==n) break;
                }
            }

            if (i<7) {  // found n-th occurence
                k = (i+new_shift) & 0x7;           
                // search next
                for (i++; i<8; i++) {
                    if (pgm_read_byte(&patterns[new_pattern].pattern[i]) == j) {
                        // next occurence
                        l = (i+new_shift) & 0x7;
                        if (k < l) {
                            set_pattern[k][l] = j;
                        } else {
                            set_pattern[l][k] = j;
                        }
                    }
                }
            }
        }
    }
}
