#include <stdlib.h>
#include <string.h>
#include <avr/io.h>
#include <util/atomic.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include "main.h"
#include "globals.h"
#include "misc.h"
#include "ws2812.h"

// read ADC in 8 bit mode

uint8_t read_ADC(uint8_t channel) {

    ADMUX = (ADMUX & ~0x7) | (channel & 0x7); 
    ADCSRA |= (1<<ADSC);
    while (ADCSRA & (1<<ADSC) );    // wait for end of conversion
    return ADCH;
}

// sample all ADC channels 16 times to get zero level

void calibrate_matrix(uint8_t data[SIZE]) {
    int i, j, tmp;

    for (i=0; i<SIZE; i++) {
        tmp = 0;
        for (j=0; j<16; j++) {
            tmp += read_ADC(i);
        }
        matrix_cal[i] = tmp / 16;
    }
}

// scan matrix

void scan_inputs(matrix_t data, int debug) {
    int i, j;

    memset(data, 0, sizeof(matrix_t));      // clear input matrix

    for (i=0; i<SIZE-1; i++) {
		// turn on scan line
        switch(i) {
            case 0: PORTC |= (1<<PC0); DDRC |= (1<<PC0); break;
            case 1: PORTC |= (1<<PC1); DDRC |= (1<<PC1); break;
            case 2: PORTC |= (1<<PC2); DDRC |= (1<<PC2); break;
            case 3: PORTC |= (1<<PC3); DDRC |= (1<<PC3); break;
            case 4: PORTC |= (1<<PC4); DDRC |= (1<<PC4); break;
            case 5: PORTC |= (1<<PC5); DDRC |= (1<<PC5); break;
            case 6: PORTB |= (1<<PB1); DDRB |= (1<<PB1); break;
            case 7: PORTB |= (1<<PB0); DDRB |= (1<<PB0); break;
        }
        _delay_ms(SAMPLE_DELAY);
        for (j=i+1; j<SIZE; j++) {
            if (debug) {    // store plain ADC data, for debug only!!!
                data[i][j] = read_ADC(j);
            } else {        // calibrate and validate input, normal operation
                if ( abs(read_ADC(j) - matrix_cal[j]) > NOISE_LEVEL ) {
                    data[i][j] = 1;
                }
            }
        }
		// turn off scan line
        switch(i) {
            case 0: DDRC &= ~(1<<PC0); PORTC &= ~(1<<PC0); break;
            case 1: DDRC &= ~(1<<PC1); PORTC &= ~(1<<PC1); break;
            case 2: DDRC &= ~(1<<PC2); PORTC &= ~(1<<PC2); break;
            case 3: DDRC &= ~(1<<PC3); PORTC &= ~(1<<PC3); break;
            case 4: DDRC &= ~(1<<PC4); PORTC &= ~(1<<PC4); break;
            case 5: DDRC &= ~(1<<PC5); PORTC &= ~(1<<PC5); break;
            case 6: DDRB &= ~(1<<PB1); PORTB &= ~(1<<PB1); break;
            case 7: DDRB &= ~(1<<PB0); PORTB &= ~(1<<PB0); break;
        }                              
    }
}

// update all 8 LEDs

void update_LEDs(led_t data) {
    int i;

    for(i=0; i < 8; i++) {
        ws2812_write(data[i], 3, &PORTD, 1<<i);
    }
} 

// set leds based on color table array

void set_all_LEDs(led_t led, PGM_P menu) {
    int i, j;

    for (i=0; i<8; i++) {
        j = pgm_read_byte(&menu[i]);
        led[i][0] = pgm_read_byte(&colors[j][0]);
        led[i][1] = pgm_read_byte(&colors[j][1]);
        led[i][2] = pgm_read_byte(&colors[j][2]);
    }
}

void set_single_LED(led_t led, uint8_t number, uint8_t color) {

    led[number][0] = pgm_read_byte(&colors[color][0]);
    led[number][1] = pgm_read_byte(&colors[color][1]);
    led[number][2] = pgm_read_byte(&colors[color][2]);
}

// count touch connections in input matrix

int count_touch(matrix_t data) {
   int i, j, cnt;

    for (i=0, cnt=0; i<SIZE-1; i++) {
        for (j=i+1; j<SIZE; j++) {
            if ( data[i][j] ) cnt++;
        }
    }
    return cnt;
}

// compare input matrix to level setpoint

int compare_matrix(matrix_t in_matrix, matrix_t set_pattern) {
    int i, j, match=1;

    for (i=0; match && i<SIZE-1; i++) {
       for (j=i+1; match && j<SIZE; j++) {
            if ( in_matrix[i][j] != set_pattern[i][j] ) {
                match = 0;
            }
        }
    }
    return match;
}

// check which menu is selected

int check_menu_select(matrix_t in_matrix) {

    if (count_touch(in_matrix) == 1) {
        if ( in_matrix[4][7] ) {
            return 0;
        }

        if ( in_matrix[0][4] ) {
            return 1;
        }

        if ( in_matrix[1][4] ) {
            return 2;
        }
    }

    return 3;
}

// generate an new random number from min to max

int rand_range(int min, int max) {

    int tmp;

    tmp = (((int32_t)rand()*(max-min+1))>>15)+min;
    return tmp;
}


void soft_uart_putc(char data) {
    unsigned int tmp, i;

    ATOMIC_BLOCK(ATOMIC_RESTORESTATE) {
        tmp = ((int)data << 1) | 0x200;

        for (i=0; i<10; i++) {
            if (tmp & 1) {
                PORTB |=  (1<<PB7);
            } else {
                PORTB &= ~(1<<PB7);
            }
            _delay_us(BIT_TIME);   
            tmp >>= 1;
        }
    }
}

void soft_uart_puts(char *data) {
    while(*data) {
        soft_uart_putc(*data++);
    }
}

void soft_uart_puts_P(PGM_P data) {
    char c;

    while( (c=pgm_read_byte(data++)) ) {
        soft_uart_putc(c);
    }
}

// generate a power pulse 100ms, 5s period
// to prevent the power bank from auto power down due to low current demand
// some need it, some don't

void power_bank_keep_alive(void) {
    static uint8_t cnt;

    cnt++;
    if (cnt == KEEP_ALIVE_PERIOD) {
        cnt = 0;
        PORTB |= (1<<PB2);
    } else {
        PORTB &= ~(1<<PB2);
    }
}
