/*
 * ws2812.S
 *
 * Created: 17.12.2017
 *  Author: Falk Brunner
 */
 
.nolist
#include <avr/io.h>	
.list

// USE_FAST_FLEXIBLE 0 : F_CPU must be 16000000, 386 Bytes flash usage
// USE_FAST_FLEXIBLE 1 : F_CPU can be 6300000-20000000, 886 bytes flash usage

#define USE_FAST_FLEXIBLE 1


#if USE_FAST_FLEXIBLE == 1

#if (F_CPU < 6300000) || (F_CPU > 20000000)
  #error F_CPU out of range! Must be 6300000-20000000!
#endif

#else

#if F_CPU != 16000000
    #error F_CPU must be 16000000. Use USE_FAST_FLEXIBLE 1 for flexible version.
#endif

#endif  // USE_FAST_FLEXIBLE == 1

.nolist

// delay of N CPU cycles using minimum code space

.macro  my_delay N
.if (\N % 3) == 1
    nop
.endif

.if (\N % 3) == 2
    nop
    nop
.endif

.if \N >= 3
    ldi     bit_cnt, \N / 3
loop\@:
    dec     bit_cnt
    brne    loop\@
.endif
.endm

// delay loop for timing tolerance test, delay for N*3 cycles

.macro  test_delay_loop N
.if \N
    ldi     bit_cnt, \N
loop\@:
    dec     bit_cnt
    brne    loop\@
.endif
.endm


/* bit pattern generator macro */

.macro bitgen
    lsl     led_data
    st      X, port_set
    brcs    MSB_is_1\@   
    ; bit is 0, minimum short pulse width is 3 F_CPU cycles
#if   F_CPU >= 18800000
    my_delay 5 
#elif F_CPU >= 16300000
    my_delay 4 
#elif F_CPU >= 13800000
    my_delay 3 
#elif F_CPU >= 11300000
    my_delay 2 
#elif F_CPU >=  8800000
    my_delay 1 
#endif
    st      X, port_clr
    ; bit is 0, minimum long pulse width is 5 F_CPU cycles
#if   F_CPU >= 19500000
    my_delay 12
#elif F_CPU >= 18300000
    my_delay 11
#elif F_CPU >= 17100000
    my_delay 10
#elif F_CPU >= 15900000
    my_delay 9
#elif F_CPU >= 14800000
    my_delay 8
#elif F_CPU >= 13600000
    my_delay 7
#elif F_CPU >= 12400000
    my_delay 6
#elif F_CPU >= 11200000
    my_delay 5
#elif F_CPU >= 10100000
    my_delay 4
#elif F_CPU >=  8900000
    my_delay 3
#elif F_CPU >=  7700000
    my_delay 2
#elif F_CPU >=  6500000
    my_delay 1
#endif
    rjmp    end\@
MSB_is_1\@:
    nop     ; equalize branch run time!
    ; bit is 1, minimum long pulse width is 5 F_CPU cycles
#if   F_CPU >= 19500000
    my_delay 12
#elif F_CPU >= 18300000
    my_delay 11
#elif F_CPU >= 17100000
    my_delay 10
#elif F_CPU >= 15900000
    my_delay 9
#elif F_CPU >= 14800000
    my_delay 8
#elif F_CPU >= 13600000
    my_delay 7
#elif F_CPU >= 12400000
    my_delay 6
#elif F_CPU >= 11200000
    my_delay 5
#elif F_CPU >= 10100000
    my_delay 4
#elif F_CPU >=  8900000
    my_delay 3
#elif F_CPU >=  7700000
    my_delay 2
#elif F_CPU >=  6500000
    my_delay 1
#endif
    st      X, port_clr
    ; bit is 1, minimum short pulse width is 3 F_CPU cycles
#if   F_CPU >= 18800000
    my_delay 5 
#elif F_CPU >= 16300000
    my_delay 4 
#elif F_CPU >= 13800000
    my_delay 3 
#elif F_CPU >= 11300000
    my_delay 2 
#elif F_CPU >=  8800000
    my_delay 1 
#endif
end\@:
.endm

.list

// register name assignment

bit_cnt    = 17 
mask       = 18
led_data   = 19
port_set   = 20
port_clr   = 21
data_cnt   = 22         ; + r23
port_ptr   = 26         ; X
led_ptr    = 30         ; Z

.extern cnt_2kHz

/*
void ws2812_write(uint8_t *p_data, uint16_t cnt, uint8_t *port, uint8_t mask);
registers                 r24/r25       r22/r23        r20/r21       r18/r19

*/

.global ws2812_write
ws2812_write:
    push    r17
    push    r26
    push    r27
    clr     r1

    lds     r17, SREG       ; restore I-Bit at the end
    push    r17

    cp      data_cnt, r1             
    cpc     data_cnt+1, r1
    brne    ws2812_w_go  
    rjmp    ws2812_w_end    ; skip function, when cnt==00
ws2812_w_go:
    movw    led_ptr, r24    ; data pointer to Z, free r24/r25
    movw    port_ptr, r20   ; port pointer to X, free r20/r21

    ld      port_set, X
    or      port_set, mask  ; port data with bit set
    com     mask
    mov     port_clr, port_set
    and     port_clr, mask  ; port data with bit clear
ws2812_w_byte_loop:
    ldi     bit_cnt, 8
    ld      led_data, Z+     ; load next data
    cli          

#if USE_FAST_FLEXIBLE == 1

; bit pattern generation using an unrolled loop
; for maximum speed at low clock rates, uses more flash memory
    bitgen      ; bit #0
    bitgen      ; bit #1
    bitgen      ; bit #2
    bitgen      ; bit #3
    bitgen      ; bit #4
    bitgen      ; bit #5
    bitgen      ; bit #6
    bitgen      ; bit #7

#else               

; classic version with bit loop, uses less flash memory
ws2812_w_bit_loop:
    st      X, port_set
    lsl     led_data
    brcs    ws2812_w_bit_is_1   ; test MSB
    ; bit is 0, short pulse 6 clocks    
    nop
    nop
    st      X, port_clr
    ; bit is 0, long pulse 14 clocks
    nop
    nop
    nop
    nop
    nop
    nop
    nop    
    rjmp    ws2812_w_bit_end
ws2812_w_bit_is_1:
    ; bit is 1, long pulse 14 clocks
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    nop
    st      X, port_clr
   ; bit is 1, short pulse 6 clocks
    nop
ws2812_w_bit_end:
    subi    bit_cnt, 1
    brne    ws2812_w_bit_loop
#endif



    ; use this additional delay loop to test for maximum tolerated time between bytes
    ; set to 0 when not used

    test_delay_loop 0

    ; enable interupts only, when LEDs can tolerate a longer break in the data signal!
    ; WS2812B can't!!!
    ;sei                 
    subi    data_cnt, 1
    sbc     data_cnt+1, r1     
    breq    ws2812_w_end
    rjmp    ws2812_w_byte_loop

ws2812_w_end:
    pop     r17
    sts     SREG, r17
    pop     r27
    pop     r26
    pop     r17
	ret







